indexing

	description: "Given INTEGER_8 input, calculates its MD5 checksum."
	standards: "Based on RFC 1321."

	limitations: "Only up to 31-bits of input is handled. But the algorithm should be easy to enhance to INTEGER_64 if so desired.."

	library: "Gobo Eiffel Pipework Library"
	copyright: "Copyright (c) 2003, Berend de Boer and others"
	license: "Eiffel Forum License v2 (see forum.txt)"
	date: "$Date: $"
	revision: "$Revision: $"

class

	PW_MD5_OUTPUT_STREAM

inherit

	PW_WRITER [INTEGER_8]
		redefine
			flush
		end

	KL_IMPORTED_INTEGER_ROUTINES
		export
			{NONE} all
		end

	KL_IMPORTED_STRING_ROUTINES
		export
			{NONE} all
		end

	KL_SHARED_PLATFORM
		export
			{NONE} all
		end


creation

	make


feature {NONE} -- Initialization

	make is
		do
			create input_buffer.make (0, 63)
			reset_state
		end


feature -- Access

	checksum: STRING
			-- The 128-byte checksum expressed as a string.

	length: INTEGER_64
			-- Length of message in bits. Includes padding of `flush' is called.

	length_without_padding: INTEGER_64
			-- Length of message in bits exclusive of any padding.
			-- Set after `flush' has been called.

feature -- Writing

	flush is
			-- Ends an MD5 message-digest operation, writing the message
			-- digest and zeroizing the context.
		do
			-- Append padding + length
			append_padding_bits
			append_length

			-- Message digest is A, B, C, D (low-order byte first)
			create checksum.make (32)
			checksum.append_string (to_hex_string (A))
			checksum.append_string (to_hex_string (B))
			checksum.append_string (to_hex_string (C))
			checksum.append_string (to_hex_string (D))

			-- Reinitialize for next round
			reset_state
		ensure then
			checksum_length_ok: checksum /= Void and then checksum.count = 32
			checksum_is_hex_string: STRING_.is_hexadecimal (checksum)
			length_reset: length = 0
		end

	put (v: INTEGER_8) is
			-- Write `v' to output stream.
		do
			-- Fill buffer `X' with 16 unsigned bytes.
			-- Signed bytes give wrong results for additions and such.
			input_buffer.put (v, j)
			if j = input_buffer.upper then
				perform_rounds
				j := input_buffer.lower
			else
				j := j + 1
			end
			length := length + 8
		ensure then
			length_incremented: length = old length + 8
		end


feature {NONE} -- MD5 operations

	append_length is
			-- A 64-bit representation of the length of the message
			-- (before the padding bits were added), is appended. Only
			-- 64-bits of the length are used.
		require
			just_enough_room_for_length: full_length = length + number_of_length_bits
		local
			k: INTEGER
			byte: INTEGER_64
		do
			from
				k := 0
			invariant
				k \\ 8 = 0
			variant
				64 - k
			until
				k = 64
			loop
				byte := length_without_padding.bit_shift_right (k) & 0xff
				put (byte.to_integer_8)
				k := k + 8
			end
		ensure
			length_is_multiple_of_512_bits: length \\ 512 = 0
		end

	append_padding_bits is
			-- Extend message so its length (in bits) is congruent to
			-- 448, modulo 512. That is, the message is extended so that
			-- it is just 64 bits shy of being a multiple of 512 bits
			-- long.
			-- Padding is always performed, even if the length fo the
			-- message is already congruent to 448 modulo 512.
		local
			bytes_to_pad: INTEGER
		do
			length_without_padding := length
			-- Write the 1 bit + seven zero bits.
			put (padding_front_byte)

			-- Fill until 8 bytes short of being a multiple of 512 bits.
			bytes_to_pad := (((full_length - length) // 8) - number_of_length_bytes).to_integer
				check
					bytes_to_pad_not_negative: bytes_to_pad >= 0
				end
			from
			variant
				bytes_to_pad
			until
				bytes_to_pad = 0
			loop
				put (0)
				bytes_to_pad := bytes_to_pad - 1
			end
		ensure
			length_without_padding_set: length_without_padding = old length
			just_enough_room_for_length: full_length = length + number_of_length_bits
		end

	full_length: INTEGER_64 is
			-- Full length of message in bits including padding and length bits.
		do
			Result := ((length // 512) + 1) * 512
			if Result - length < number_of_length_bits then
				Result := Result + 512
			end
		ensure
			length_is_multiple_of_512_bits: Result \\ 512 = 0
			always_room_for_length_bits: Result - length >= number_of_length_bits
		end

	initialize_md5_buffer is
			-- Step 3 in RFC 1321 implementation.
		do
			A := A_initial_value
			B := B_initial_value
			C := C_initial_value
			D := D_initial_value
		ensure
			definition:
				A = A_initial_value and then
				B = B_initial_value and then
				C = C_initial_value and then
				D = D_initial_value
		end

	number_of_length_bits: INTEGER is 64
			-- Number of bits required for length of message.

	number_of_length_bytes: INTEGER is 8
			-- Number of bytes required for length of message.

	perform_rounds is
			-- The rounds mentioned in the algorithm in Step 4.
			-- Thanks to the Nenie project for a few ideas.
		local
			AA, BB, CC, DD: INTEGER
			block_0, block_1, block_2, block_3,
			block_4, block_5, block_6, block_7,
			block_8, block_9, block_10, block_11,
			block_12, block_13, block_14, block_15: INTEGER
		do
			-- Save A as AA, B as BB, C as CC, and D as DD
			AA := A
			BB := B
			CC := C
			DD := D

			-- Compress 64 8-bits into 16 32-bits.
			block_0 := input_buffer_32_bits (0)
			block_1 := input_buffer_32_bits (4)
			block_2 := input_buffer_32_bits (8)
			block_3 := input_buffer_32_bits (12)
			block_4 := input_buffer_32_bits (16)
			block_5 := input_buffer_32_bits (20)
			block_6 := input_buffer_32_bits (24)
			block_7 := input_buffer_32_bits (28)
			block_8 := input_buffer_32_bits (32)
			block_9 := input_buffer_32_bits (36)
			block_10 := input_buffer_32_bits (40)
			block_11 := input_buffer_32_bits (44)
			block_12 := input_buffer_32_bits (48)
			block_13 := input_buffer_32_bits (52)
			block_14 := input_buffer_32_bits (56)
			block_15 := input_buffer_32_bits (60)

			-- Round 1
			A := FF (A, B, C, D, block_0, 7, sin_01)
			D := FF (D, A, B, C, block_1, 12, sin_02)
			C := FF (C, D, A, B, block_2, 17, sin_03)
			B := FF (B, C, D, A, block_3, 22, sin_04)
			A := FF (A, B, C, D, block_4, 7, sin_05)
			D := FF (D, A, B, C, block_5, 12, sin_06)
			C := FF (C, D, A, B, block_6, 17, sin_07)
			B := FF (B, C, D, A, block_7, 22, sin_08)
			A := FF (A, B, C, D, block_8, 7, sin_09)
			D := FF (D, A, B, C, block_9, 12, sin_10)
			C := FF (C, D, A, B, block_10, 17, sin_11)
			B := FF (B, C, D, A, block_11, 22, sin_12)
			A := FF (A, B, C, D, block_12, 7, sin_13)
			D := FF (D, A, B, C, block_13, 12, sin_14)
			C := FF (C, D, A, B, block_14, 17, sin_15)
			B := FF (B, C, D, A, block_15, 22, sin_16)

			-- Round 2
			A := GG (A, B, C, D, block_1, 5, sin_17)
			D := GG (D, A, B, C, block_6, 9, sin_18)
			C := GG (C, D, A, B, block_11, 14, sin_19)
			B := GG (B, C, D, A, block_0, 20, sin_20)
			A := GG (A, B, C, D, block_5, 5, sin_21)
			D := GG (D, A, B, C, block_10, 9, sin_22)
			C := GG (C, D, A, B, block_15, 14, sin_23)
			B := GG (B, C, D, A, block_4, 20, sin_24)
			A := GG (A, B, C, D, block_9, 5, sin_25)
			D := GG (D, A, B, C, block_14, 9, sin_26)
			C := GG (C, D, A, B, block_3, 14, sin_27)
			B := GG (B, C, D, A, block_8, 20, sin_28)
			A := GG (A, B, C, D, block_13, 5, sin_29)
			D := GG (D, A, B, C, block_2, 9, sin_30)
			C := GG (C, D, A, B, block_7, 14, sin_31)
			B := GG (B, C, D, A, block_12, 20, sin_32)

			-- Round 3
			A := HH (A, B, C, D, block_5, 4, sin_33)
			D := HH (D, A, B, C, block_8, 11, sin_34)
			C := HH (C, D, A, B, block_11, 16, sin_35)
			B := HH (B, C, D, A, block_14, 23, sin_36)
			A := HH (A, B, C, D, block_1, 4, sin_37)
			D := HH (D, A, B, C, block_4, 11, sin_38)
			C := HH (C, D, A, B, block_7, 16, sin_39)
			B := HH (B, C, D, A, block_10, 23, sin_40)
			A := HH (A, B, C, D, block_13, 4, sin_41)
			D := HH (D, A, B, C, block_0, 11, sin_42)
			C := HH (C, D, A, B, block_3, 16, sin_43)
			B := HH (B, C, D, A, block_6, 23, sin_44)
			A := HH (A, B, C, D, block_9, 4, sin_45)
			D := HH (D, A, B, C, block_12, 11, sin_46)
			C := HH (C, D, A, B, block_15, 16, sin_47)
			B := HH (B, C, D, A, block_2, 23, sin_48)

			-- Round 4
			A := II (A, B, C, D, block_0, 6, sin_49)
			D := II (D, A, B, C, block_7, 10, sin_50)
			C := II (C, D, A, B, block_14, 15, sin_51)
			B := II (B, C, D, A, block_5, 21, sin_52)
			A := II (A, B, C, D, block_12, 6, sin_53)
			D := II (D, A, B, C, block_3, 10, sin_54)
			C := II (C, D, A, B, block_10, 15, sin_55)
			B := II (B, C, D, A, block_1, 21, sin_56)
			A := II (A, B, C, D, block_8, 6, sin_57)
			D := II (D, A, B, C, block_15, 10, sin_58)
			C := II (C, D, A, B, block_6, 15, sin_59)
			B := II (B, C, D, A, block_13, 21, sin_60)
			A := II (A, B, C, D, block_4, 6, sin_61)
			D := II (D, A, B, C, block_11, 10, sin_62)
			C := II (C, D, A, B, block_2, 15, sin_63)
			B := II (B, C, D, A, block_9, 21, sin_64)

			-- Increment each of the four registers by the value it had
			-- before this block was started.
			A := A + AA
			B := B + BB
			C := C+  CC
			D := D + DD
		end

	reset_state is
			-- Reset state so digest can be calculated for a new message.
		do
			initialize_md5_buffer
			length := 0
			j := input_buffer.lower
		ensure
			definition:
				A = A_initial_value and then
				B = B_initial_value and then
				C = C_initial_value and then
				D = D_initial_value
			j_reset: j = input_buffer.lower
			length_reset: length = 0
		end


feature {NONE} -- MD5 functions

	FF (aa, bb, cc, dd, x: INTEGER; s: INTEGER_8; ac: INTEGER): INTEGER is
			-- Let [abcd x s i] denote the operation
			-- f() = b + ((a + F(b,c,d) + x + T[i]) unsigned_rotate_left s)
		require
			shift_positive: s > 0
		do
			Result := aa + F (bb, cc, dd) + x + ac
			Result := rotate_left (Result, s)
			Result := Result + bb
		end

	GG (aa, bb, cc, dd, x: INTEGER; s: INTEGER_8; ac: INTEGER): INTEGER is
			-- Let [abcd k s i] denote the operation
			-- f() = b + ((a + G(b,c,d) + x + T[i]) unsigned_rotate_left s)
		require
			shift_positive: s > 0
		do
			Result := aa + G (bb, cc, dd) + x + ac
			Result := rotate_left (Result, s)
			Result := Result + bb
		end

	HH (aa, bb, cc, dd, x: INTEGER; s: INTEGER_8; ac: INTEGER): INTEGER is
			-- Let [abcd k s i] denote the operation
			-- f() = b + ((a + H(b,c,d) + x + T[i]) unsigned_rotate_left s)
		require
			shift_positive: s > 0
		do
			Result := aa + H (bb, cc, dd) + x + ac
			Result := rotate_left (Result, s)
			Result := Result + bb
		end

	II (aa, bb, cc, dd, x: INTEGER; s: INTEGER_8; ac: INTEGER): INTEGER is
			-- Let [abcd k s i] denote the operation
			-- f() = b + ((a + I(b,c,d) + x + T[i]) unsigned_rotate_left s)
		require
			shift_positive: s > 0
		do
			Result := aa + I (bb, cc, dd) + x + ac
			Result := rotate_left (Result, s)
			Result := Result + bb
		end


feature {NONE} -- MD5 primitives

	rotate_left (value: INTEGER; n: INTEGER_8): INTEGER is
		require
			n_positive: n > 0
		do
			--Result := value.bit_shift_left (n) | value.bit_shift_right_unsigned (32 - n)
			Result := value.bit_shift_left (n) | bit_shift_right_unsigned (value, 32 - n)
		end

	bit_shift_right_unsigned (value: INTEGER; n: INTEGER): INTEGER is
			-- Shift `value' by `n' positions right (sign bit not copied) bits
			-- falling off the end are lost.
			-- Routine here because ISE does not have an unsigned shift.
			-- Routine thanks to the Nenie project.
		require
			n_positive: n > 0
		do
			if value < 0 then
				-- bit_shift_right does sign extension, so shift by one,
				-- clear top bit, and do remainder bits on positive
				-- result.
				Result := value.bit_shift_right (1)
				Result := Result.bit_and ((1).bit_shift_left (31).bit_not)
				Result := Result.bit_shift_right (n -  1)
			else
					check
						sign_bit_clear: value >= 0
					end
				Result := value.bit_shift_right (n)
			end
		end

	F (xx, yy, zz: INTEGER): INTEGER is
			-- F(X,Y,Z) = XY v not(X) Z.
		do
			Result := (xx & yy) | (xx.bit_not & zz)
		end

	G (xx, yy, zz: INTEGER): INTEGER is
			-- G(X,Y,Z) = XZ v Y not(Z).
		do
			Result := (xx & zz) | (yy & zz.bit_not)
		end

	H (xx, yy, zz: INTEGER): INTEGER is
			-- H(X,Y,Z) = X xor Y xor Z.
		do
			Result := xx.bit_xor (yy.bit_xor (zz))
		end

	I (xx, yy, zz: INTEGER): INTEGER is
			-- I(X,Y,Z) = Y xor (X | not(Z))
		do
			Result := yy.bit_xor (xx | zz.bit_not)
		end


feature {NONE} -- MD5 registers

	A, B, C, D: INTEGER
			-- 32-bit registers.

	A_initial_value: INTEGER is 1732584193
			-- 01 23 45 67 (low-order byte first)

	B_initial_value: INTEGER is -271733879
		-- 89 ab cd ef (low-order byte first)

	C_initial_value: INTEGER is -1732584194
		-- fe dc ba 98 (low-order byte first)

	D_initial_value: INTEGER is 271733878
		-- 76 54 32 10 (low-order byte first)

	input_buffer: ARRAY [INTEGER_8]
			-- Input buffer of 8-bit values.

	input_buffer_32_bits (index: INTEGER): INTEGER is
			-- Return 32-bits from `input_buffer' starting at byte `index'.
		require
			valid_index: index >= input_buffer.lower and index <= input_buffer.upper
		do
			Result :=
				unsigned_item (index) |
				unsigned_item (index + 1).bit_shift_left (8) |
				unsigned_item (index + 2).bit_shift_left (16) |
				unsigned_item (index + 3).bit_shift_left (24)
		end

	j: INTEGER
			-- Index into `input_buffer' where to store the next incoming byte.

	padding_front_byte: INTEGER_8 is
		once
			Result := Result.Min_value
		ensure
			Result.to_integer = -128
		end

	unsigned_item (index: INTEGER): INTEGER is
			-- Return an unsigned item from `input_buffer'.
		require
			valid_index: index >= input_buffer.lower and index <= input_buffer.upper
		do
			Result := input_buffer.item (index)
			if Result < 0 then
				Result := 256 + Result
			end
		ensure
			unsigned: Result >= 0
		end


feature {NONE} -- MD5 sin table: T[i] = int part of 4294967296 * (abs(sin(i)))

	sin_01: INTEGER is -680876936
	sin_02: INTEGER is -389564586
	sin_03: INTEGER is 606105819
	sin_04: INTEGER is -1044525330
	sin_05: INTEGER is -176418897
	sin_06: INTEGER is 1200080426
	sin_07: INTEGER is -1473231341
	sin_08: INTEGER is -45705983
	sin_09: INTEGER is 1770035416
	sin_10: INTEGER is -1958414417
	sin_11: INTEGER is -42063
	sin_12: INTEGER is -1990404162
	sin_13: INTEGER is 1804603682
	sin_14: INTEGER is -40341101
	sin_15: INTEGER is -1502002290
	sin_16: INTEGER is 1236535329
	sin_17: INTEGER is -165796510
	sin_18: INTEGER is -1069501632
	sin_19: INTEGER is 643717713
	sin_20: INTEGER is -373897302
	sin_21: INTEGER is -701558691
	sin_22: INTEGER is 38016083
	sin_23: INTEGER is -660478335
	sin_24: INTEGER is -405537848
	sin_25: INTEGER is 568446438
	sin_26: INTEGER is -1019803690
	sin_27: INTEGER is -187363961
	sin_28: INTEGER is 1163531501
	sin_29: INTEGER is -1444681467
	sin_30: INTEGER is -51403784
	sin_31: INTEGER is 1735328473
	sin_32: INTEGER is -1926607734
	sin_33: INTEGER is -378558
	sin_34: INTEGER is -2022574463
	sin_35: INTEGER is 1839030562
	sin_36: INTEGER is -35309556
	sin_37: INTEGER is -1530992060
	sin_38: INTEGER is 1272893353
	sin_39: INTEGER is -155497632
	sin_40: INTEGER is -1094730640
	sin_41: INTEGER is 681279174
	sin_42: INTEGER is -358537222
	sin_43: INTEGER is -722521979
	sin_44: INTEGER is 76029189
	sin_45: INTEGER is -640364487
	sin_46: INTEGER is -421815835
	sin_47: INTEGER is 530742520
	sin_48: INTEGER is -995338651
	sin_49: INTEGER is -198630844
	sin_50: INTEGER is 1126891415
	sin_51: INTEGER is -1416354905
	sin_52: INTEGER is -57434055
	sin_53: INTEGER is 1700485571
	sin_54: INTEGER is -1894986606
	sin_55: INTEGER is -1051523
	sin_56: INTEGER is -2054922799
	sin_57: INTEGER is 1873313359
	sin_58: INTEGER is -30611744
	sin_59: INTEGER is -1560198380
	sin_60: INTEGER is 1309151649
	sin_61: INTEGER is -145523070
	sin_62: INTEGER is -1120210379
	sin_63: INTEGER is 718787259
	sin_64: INTEGER is -343485551


feature {NONE} -- Implementation

	hex_digits: STRING is "0123456789abcdef"

	reverse_positions: ARRAY [INTEGER] is
		once
			Result := <<7, 8, 5, 6, 3, 4, 1, 2>>
		end

	to_hex_string (item: INTEGER): STRING is
			-- Convert `item' into an hexadecimal string.
			-- Low-order byte is listed first.
		local
			k, val: INTEGER
			a_digit: INTEGER
		do
			from
				k := Platform.Integer_bits // 4
				create Result.make_filled ('0', k)
				val := item
			variant
				k
			until
				k = 0
			loop
				a_digit := (val & 0x0F)
				Result.put (hex_digits.item (a_digit + 1), reverse_positions.item (k))
				val := val |>> 4
				k := k - 1
			end
		ensure
			Result_not_void: Result /= Void
			Result_valid_count: Result.count = Platform.Integer_bits // 4
			is_hex_string: STRING_.is_hexadecimal (Result)
		end


invariant

	input_buffer_has_64_bytes: input_buffer /= Void and then input_buffer.count = 64
	j_within_bounds_of_input_buffer: j >= input_buffer.lower and j <= input_buffer.upper
	length_positive: length >= 0
	length_without_padding_positive: length_without_padding >= 0


end
